// ============================================================================
//
// Source7.c
// CBP 24-03-22
// Timing Tests for good and bad memory access 
// MIT p127
// Use PROBSIZE (array length) as independent variable
// Could use this as an argument and use malloc.
// Part of NoC Multi-Processing Workshops
// Status: Release
//
// ============================================================================

#include <stdio.h>
#include <stdlib.h>
#include <time.h>		// Needed for clock_t
#include <string.h>
#include <omp.h>

#define NR_ROWS 20000
#define NR_COLS 2000
#define DES_NR_THREADS 16
#define REPEAT_NUM 10

int threadNr;
int x;
int y;
clock_t t0, t1, t2;
double time1, time2;
int nrProcs;
int nrThreads;
double wtime;
double* a;
double sum;
int nrRows, nrCols;

int main(int argc, char** argv) {

	printf("============== CBP NoC 2022 Multi-Processing Workshops ============\n\n");
	printf("---------------------------Activity 7 -----------------------------\n");

	printf("Request to system for coresand threads\n");
	system("WMIC CPU Get NumberofCores,NumberOfLogicalProcessors");

	nrProcs = omp_get_num_procs();
	printf("OMP says nr of threads on this machine = %d\n", nrProcs);

	printf("Nr Threads requested = %d\n", DES_NR_THREADS);

	printf("REPEAT_NUM = %d\n", REPEAT_NUM);

	omp_set_dynamic(0);     // Explicitly disable dynamic teams
	omp_set_num_threads(DES_NR_THREADS);

	printf("Nr Threads requested = %d\n", DES_NR_THREADS);

	a = (double*)malloc(NR_ROWS * NR_COLS * sizeof(double));

	nrRows = NR_ROWS;
	nrCols = NR_COLS;

	for (long i = 0; i < NR_ROWS * NR_COLS; i++) {
		//a[i] = 1.0;
		*(a + i) = 1.0;
	}

	t0 = clock();
	wtime = omp_get_wtime();

	t1 = clock();

	for (int repeat = 0; repeat < REPEAT_NUM; repeat++) {
		sum = 0;
		#pragma omp parallel 
		{

			#pragma omp for
			// First variant ------------------------------
			for (int row = 0; row < nrRows; row++) {
				for (int col = 0; col < nrCols; col++) {
					sum += a[row * nrCols + col];
				}
			}
			// Second variant ----------------------------
			//for (int col = 0; col < nrCols; col++) {
			//	for (int row = 0; row < nrRows; row++) {
			//		sum += a[row * nrCols + col];
			//	}
			//}
		}
	}

	printf("\nSum = %f\n", sum);
	t2 = clock();
	time2 = (double)(t2)-(double)(t1);
	printf("clock() Time = %f (ms)\n", time2);

	wtime = omp_get_wtime() - wtime;
	printf("wtime = %f (mSec)\n", 1000 * wtime);

	free(a);

}